from crispy_forms_gds.helper import FormHelper
from crispy_forms_gds.layout import HTML, Field, Fieldset, Hidden, Layout, Size, Submit
from django.forms import ModelForm, RadioSelect

from django_feedback_govuk.models import BaseFeedback, Feedback, SatisfactionOptions
from django_feedback_govuk.settings import dfg_settings


SUBMIT_BUTTON = Submit("submit", "Send feedback")


class BaseFeedbackForm(ModelForm):
    class Meta:
        model = BaseFeedback
        fields = ["submitter"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        submitter = self.initial["submitter"]
        submitter_id = str(submitter.id) if submitter.id else ""

        self.fields["submitter"].required = False

        self.helper = FormHelper()
        self.helper.layout = Layout(
            Hidden("submitter", submitter_id),
            SUBMIT_BUTTON,
        )


class FeedbackForm(BaseFeedbackForm):
    class Meta:
        model = Feedback
        fields = ["satisfaction", "comment", "submitter"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields["satisfaction"].label = ""
        self.fields["satisfaction"].required = True
        self.fields["satisfaction"].widget = RadioSelect()
        self.fields["satisfaction"].choices = SatisfactionOptions.choices
        self.fields["comment"].label = ""

        self.helper.layout.remove(SUBMIT_BUTTON)
        self.helper.layout.append(
            Fieldset(
                Field.radios(
                    "satisfaction",
                    template="django_feedback_govuk/widgets/star_rating/star_rating.html",
                ),
                legend=dfg_settings.COPY_FIELD_SATISFACTION_LEGEND,
                legend_size=Size.MEDIUM,
            )
        )
        self.helper.layout.append(
            Fieldset(
                HTML(
                    f"<p class='govuk-hint'>{dfg_settings.COPY_FIELD_COMMENT_HINT}</p>"
                ),
                Field("comment"),
                legend=dfg_settings.COPY_FIELD_COMMENT_LEGEND,
                legend_size=Size.MEDIUM,
            )
        )
        self.helper.layout.append(SUBMIT_BUTTON)



from typing import Dict, List

from django.conf import settings


DEFAULT_FEEDBACK_ID = "default"

DEFAULTS = {
    "SERVICE_NAME": "Example service",
    "FEEDBACK_NOTIFICATION_EMAIL_TEMPLATE_ID": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
    "FEEDBACK_NOTIFICATION_EMAIL_RECIPIENTS": [
        "email@example.com",
    ],
    "COPY": {
        "SUBMIT_TITLE": "Give feedback on {{ service_name }}",
        "CONFIRM_TITLE": "Feedback submitted",
        "CONFIRM_BODY": "Thank you for submitting your feedback.",
        "FIELD_SATISFACTION_LEGEND": "Overall, how did you feel about the service you received today?",
        "FIELD_COMMENT_LEGEND": "How could we improve this service?",
        "FIELD_COMMENT_HINT": "Do not include any personal or financial information, for example your National Insurance or credit card numbers.",
    },
    "FEEDBACK_FORMS": {
        DEFAULT_FEEDBACK_ID: {
            "model": "django_feedback_govuk.models.Feedback",
            "form": "django_feedback_govuk.forms.FeedbackForm",
            "view": "django_feedback_govuk.views.FeedbackView",
        },
    },
}


class DjangoFeedbackGovUKSettings:
    SERVICE_NAME: str
    FEEDBACK_NOTIFICATION_EMAIL_TEMPLATE_ID: str
    FEEDBACK_NOTIFICATION_EMAIL_RECIPIENTS: List[str]
    COPY_SUBMIT_TITLE: str
    COPY_CONFIRM_TITLE: str
    COPY_CONFIRM_BODY: str
    COPY_FIELD_SATISFACTION_LEGEND: str
    COPY_FIELD_COMMENT_LEGEND: str
    COPY_FIELD_COMMENT_HINT: str
    FEEDBACK_FORMS: Dict[str, Dict[str, str]]

    def __getattr__(self, attr):
        django_settings = getattr(settings, "DJANGO_FEEDBACK_GOVUK", {})

        # Get COPY values
        if attr.startswith("COPY_"):
            copy_key = attr[5:]
            value = django_settings.get("COPY", {}).get(copy_key)
            if value:
                # Return the value from user settings
                return value
            # Return the value from defaults
            return DEFAULTS["COPY"][copy_key]

        if attr in django_settings:
            # Return the value from user settings
            return django_settings[attr]

        default_value = DEFAULTS.get(attr, None)
        if default_value is None and attr not in DEFAULTS:
            raise AttributeError(f"No value set for DJANGO_FEEDBACK_GOVUK['{attr}']")
        # Return the value from defaults
        return default_value


dfg_settings = DjangoFeedbackGovUKSettings()



from django import template
from django.utils.module_loading import import_string

from django_feedback_govuk.settings import DEFAULT_FEEDBACK_ID, dfg_settings


register = template.Library()


@register.inclusion_tag(
    "django_feedback_govuk/partials/submit.html", takes_context=True
)
def feedback_submit(context, form_id: str = DEFAULT_FEEDBACK_ID):
    feedback_config = dfg_settings.FEEDBACK_FORMS[form_id]
    feedback_form = import_string(feedback_config["form"])
    if "form" in context:
        form = context["form"]
    else:
        initial = {}
        initial["submitter"] = context.request.user

        form = feedback_form(initial=initial)

    new_context = {
        "form": form,
        "form_id": form_id,
        "service_name": dfg_settings.SERVICE_NAME,
        "submit_title": dfg_settings.COPY_SUBMIT_TITLE,
    }
    return new_context


@register.inclusion_tag(
    "django_feedback_govuk/partials/confirm.html", takes_context=True
)
def feedback_confirm(context):
    new_context = {
        "service_name": dfg_settings.SERVICE_NAME,
        "confirm_title": dfg_settings.COPY_CONFIRM_TITLE,
        "confirm_body": dfg_settings.COPY_CONFIRM_BODY,
    }
    return new_context


@register.inclusion_tag(
    "django_feedback_govuk/partials/listing.html", takes_context=True
)
def feedback_listing(context):
    return context


@register.inclusion_tag(
    "django_feedback_govuk/partials/submitted.html", takes_context=True
)
def feedback_submitted(context):
    return context


@register.filter()
def get_feedback_value(object, field_name):
    get_f_display = f"get_{field_name}_display"
    if value := getattr(object, get_f_display, None):
        return value()
    return getattr(object, field_name)


@register.filter()
def get_feedback_form_label(form, field_name):
    field = form.fields.get(field_name, None)
    if field and field.label:
        return field.label
    return field_name.replace("_", " ").capitalize()


@register.filter()
def get_elided_page_range(page):
    return page.paginator.get_elided_page_range(page.number, on_each_side=1, on_ends=1)


@register.simple_tag(takes_context=True)
def get_pagination_url(context, page):
    request = context["request"]

    query_params = request.GET.copy()
    query_params["page"] = page

    return "?" + query_params.urlencode()



from django.contrib.auth import get_user_model
from django.db import models
from django.db.models.enums import TextChoices


class SatisfactionOptions(TextChoices):
    """
    Enum for the satisfaction options.
    """

    VERY_DISSATISFIED = "very_dissatisfied", "Very dissatisfied"
    DISSATISFIED = "dissatisfied", "Dissatisfied"
    NEUTRAL = "neutral", "Neither satisfied or dissatisfied"
    SATISFIED = "satisfied", "Satisfied"
    VERY_SATISFIED = "very_satisfied", "Very satisfied"


class BaseFeedback(models.Model):
    submitter = models.ForeignKey(get_user_model(), on_delete=models.CASCADE, null=True)
    submitted_at = models.DateTimeField(null=True, auto_now_add=True)

    class Meta:
        verbose_name = "Feedback Submission"
        verbose_name_plural = "Feedback Submissions"


class Feedback(BaseFeedback):
    satisfaction = models.CharField(max_length=30, choices=SatisfactionOptions.choices)
    comment = models.TextField(blank=True)



from django.urls import path

from . import views


urlpatterns = [
    path("submit/", views.FeedbackView.as_view(), name="feedback-submit"),
    path("submit/<form_id>/", views.get_feedback_view, name="custom-feedback-submit"),
    path("confirm/", views.feedback_confirm, name="feedback-confirm"),
    path(
        "submitted/",
        views.SubmittedFeedback.as_view(),
        name="submitted-feedback",
    ),
    path(
        "submitted/<form_id>/",
        views.FeedbackListingView.as_view(),
        name="feedback-listing",
    ),
    # path("submit-ajax", views.AjaxFeedbackView.as_view(), name="feedback-submit-ajax"),
    # path("confirm-ajax/", views.feedback_confirm_ajax, name="feedback-confirm-ajax"),
]



import os

import django.utils.crypto


TESTS_PATH = os.path.dirname(os.path.abspath(__file__))

SECRET_KEY = django.utils.crypto.get_random_string(50)

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "crispy_forms",
    "crispy_forms_gds",
    "django_feedback_govuk",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": "db.sqlite3"}}

MEDIA_ROOT = os.path.join(TESTS_PATH, "media")

STATIC_ROOT = os.path.join(TESTS_PATH, "static")

ROOT_URLCONF = "django_feedback_govuk.tests.urls"

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# Crispy forms
CRISPY_ALLOWED_TEMPLATE_PACKS = ["gds"]
CRISPY_TEMPLATE_PACK = "gds"

# Django Feedback GovUK
DJANGO_FEEDBACK_GOVUK = {
    "SERVICE_NAME": "Example Project",
    "FEEDBACK_NOTIFICATION_EMAIL_TEMPLATE_ID": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
    "FEEDBACK_NOTIFICATION_EMAIL_RECIPIENTS": ["email@example.com"],
}



from typing import Dict

from django.conf import settings
from notifications_python_client.notifications import NotificationsAPIClient

from .settings import dfg_settings


def email(personalisation: Dict):
    notification_client = NotificationsAPIClient(
        settings.GOVUK_NOTIFY_API_KEY,
    )
    email_addresses = dfg_settings.FEEDBACK_NOTIFICATION_EMAIL_RECIPIENTS

    if len(email_addresses) < 1:
        raise Exception("No feedback recipients configured")

    for email_address in email_addresses:
        message_response = notification_client.send_email_notification(
            email_address=email_address,
            template_id=dfg_settings.FEEDBACK_NOTIFICATION_EMAIL_TEMPLATE_ID,
            personalisation=personalisation,
        )

    return message_response



# Generated by Django 3.2.20 on 2023-08-04 08:20

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


def migrate_from_old_to_new_feedback(apps, schema_editor):
    OldFeedback = apps.get_model("django_feedback_govuk", "OldFeedback")
    Feedback = apps.get_model("django_feedback_govuk", "Feedback")

    for old_feedback in OldFeedback.objects.all():
        new_feedback = Feedback.objects.create(
            submitter=old_feedback.submitter,
            submitted_at=old_feedback.submitted_at,
            satisfaction=old_feedback.satisfaction,
            comment=old_feedback.comment,
        )
        new_feedback.save()


def migrate_from_new_to_old_feedback(apps, schema_editor):
    OldFeedback = apps.get_model("django_feedback_govuk", "OldFeedback")
    Feedback = apps.get_model("django_feedback_govuk", "Feedback")

    for new_feedback in Feedback.objects.all():
        old_feedback = OldFeedback.objects.create(
            submitter=new_feedback.submitter,
            submitted_at=new_feedback.submitted_at,
            satisfaction=new_feedback.satisfaction,
            comment=new_feedback.comment,
        )
        old_feedback.save()


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("django_feedback_govuk", "0002_alter_feedback_submitter"),
    ]

    operations = [
        migrations.CreateModel(
            name="BaseFeedback",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("submitted_at", models.DateTimeField(auto_now_add=True, null=True)),
                (
                    "submitter",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AlterModelOptions(
            options={
                "verbose_name": "Feedback Submission",
                "verbose_name_plural": "Feedback Submissions",
            },
            name="basefeedback",
        ),
        migrations.RenameModel(
            old_name="Feedback",
            new_name="OldFeedback",
        ),
        migrations.CreateModel(
            name="Feedback",
            fields=[
                (
                    "basefeedback_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_feedback_govuk.basefeedback",
                    ),
                ),
                (
                    "satisfaction",
                    models.CharField(
                        choices=[
                            ("very_dissatisfied", "Very dissatisfied"),
                            ("dissatisfied", "Dissatisfied"),
                            ("neutral", "Neither satisfied or dissatisfied"),
                            ("satisfied", "Satisfied"),
                            ("very_satisfied", "Very satisfied"),
                        ],
                        max_length=30,
                    ),
                ),
                ("comment", models.TextField(blank=True)),
            ],
            bases=("django_feedback_govuk.basefeedback",),
        ),
        migrations.RunPython(
            migrate_from_old_to_new_feedback,
            migrate_from_new_to_old_feedback,
        ),
        migrations.DeleteModel(
            name="OldFeedback",
        ),
    ]



from typing import Any, Dict

from django import http
from django.contrib.auth.mixins import UserPassesTestMixin
from django.http.response import HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse, reverse_lazy
from django.utils.module_loading import import_string
from django.views.generic import FormView, ListView, TemplateView

from django_feedback_govuk import notify
from django_feedback_govuk.models import BaseFeedback
from django_feedback_govuk.settings import DEFAULT_FEEDBACK_ID, dfg_settings


class FeedbackView(FormView):
    template_name = "django_feedback_govuk/templates/submit.html"
    success_url = reverse_lazy("feedback-confirm")

    def dispatch(
        self, request: http.HttpRequest, *args: Any, **kwargs: Any
    ) -> http.HttpResponse:
        self.form_id = kwargs.get("form_id", DEFAULT_FEEDBACK_ID)

        try:
            self.feedback_config = dfg_settings.FEEDBACK_FORMS[self.form_id]
        except KeyError:
            raise ValueError(f"Unknown feedback form ID: {self.form_id}")
        return super().dispatch(request, *args, **kwargs)

    def get_form_class(self):
        return import_string(self.feedback_config["form"])

    def get_initial(self):
        initial = super().get_initial()
        initial["submitter"] = self.request.user
        return initial

    def form_valid(self, form):
        form.save()
        feedback_listing_path = reverse(
            "feedback-listing",
            kwargs={"form_id": self.form_id},
        )
        # Send an email to inform the team of the feedback
        notify.email(
            personalisation={
                "feedback_url": self.request.build_absolute_uri(feedback_listing_path),
            },
        )
        return super().form_valid(form)

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context["form_id"] = self.form_id
        return context


def get_feedback_view(request, *args, **kwargs):
    form_id = kwargs.get("form_id", DEFAULT_FEEDBACK_ID)
    feedback_config = dfg_settings.FEEDBACK_FORMS[form_id]
    feedback_view: FormView = import_string(feedback_config["view"])
    return feedback_view.as_view()(request, *args, **kwargs)


def feedback_confirm(request):
    return render(request, "django_feedback_govuk/templates/confirm.html")


class UserCanViewFeedback(UserPassesTestMixin):
    def test_func(self):
        return self.request.user.has_perm("django_feedback_govuk.view_feedback")


class SubmittedFeedback(UserCanViewFeedback, TemplateView):
    template_name = "django_feedback_govuk/templates/submitted.html"

    def dispatch(
        self, request: http.HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponse:
        self.feedback_forms: Dict[str, str] = {}

        for feedback_form_id in dfg_settings.FEEDBACK_FORMS:
            feedback_config = dfg_settings.FEEDBACK_FORMS.get(feedback_form_id)
            feedback_model = import_string(feedback_config["model"])
            self.feedback_forms[feedback_form_id] = {
                "submission_count": feedback_model.objects.all().count()
            }

        if len(self.feedback_forms) == 1:
            form_id = list(self.feedback_forms.keys())[0]
            return redirect(
                reverse(
                    "feedback-listing",
                    kwargs={
                        "form_id": form_id,
                    },
                )
            )

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)

        context["feedback_forms"] = self.feedback_forms

        return context


class FeedbackListingView(UserCanViewFeedback, ListView):
    template_name = "django_feedback_govuk/templates/listing.html"
    model = BaseFeedback
    paginate_by = 5

    def dispatch(
        self, request: http.HttpRequest, *args: Any, **kwargs: Any
    ) -> http.HttpResponse:
        self.form_id = kwargs.get("form_id", DEFAULT_FEEDBACK_ID)

        feedback_config = dfg_settings.FEEDBACK_FORMS[self.form_id]
        self.feedback_form = import_string(feedback_config["form"])
        self.feedback_model = import_string(feedback_config["model"])

        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.feedback_model.objects.all().order_by("-submitted_at")

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)

        form = self.feedback_form(
            initial={
                "submitter": self.request.user,
            },
        )
        form_fields = list(form.fields)
        # Move 'submitter' to the end.
        form_fields.remove("submitter")
        form_fields.append("submitter")

        context.update(
            form_id=self.form_id,
            form=form,
            fields=form_fields,
            model=self.feedback_model,
            hide_back_button=len(dfg_settings.FEEDBACK_FORMS) == 1,
        )
        return context


# #
# # AJAX supporting views - e.g. using HTMX to render without pageloads
# #

# class AjaxFeedbackView(FeedbackView):
#     template_name = "django_feedback_govuk/partials/submit.html"
#     success_url = reverse_lazy("feedback-confirm-ajax")


# def feedback_confirm_ajax(request):
#     return render(request, "django_feedback_govuk/partials/confirm.html")



# Generated by Django 4.1.10 on 2023-08-04 14:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_feedback_govuk', '0003_restructure_feedback'),
    ]

    operations = [
        migrations.AlterField(
            model_name='basefeedback',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



# Generated by Django 4.1.7 on 2023-03-14 12:55

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Feedback",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "satisfaction",
                    models.CharField(
                        choices=[
                            ("very_dissatisfied", "Very dissatisfied"),
                            ("dissatisfied", "Dissatisfied"),
                            ("neutral", "Neither satisfied or dissatisfied"),
                            ("satisfied", "Satisfied"),
                            ("very_satisfied", "Very satisfied"),
                        ],
                        max_length=30,
                    ),
                ),
                ("comment", models.TextField(blank=True)),
                (
                    "submitter",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("submitted_at", models.DateTimeField(auto_now_add=True, null=True)),
            ],
            options={
                "verbose_name": "Feedback Submission",
                "verbose_name_plural": "Feedback Submissions",
            },
        ),
    ]



from django.contrib.auth.models import Permission
from django.test import TestCase
from django.urls import reverse

from django_feedback_govuk.tests.factories import UserFactory


class TestFeedbackView(TestCase):
    view_name = "feedback-submit"

    def setUp(self):
        self.view_url = reverse(self.view_name)
        self.user = UserFactory()

    def test_feedback_view_get(self):
        self.client.force_login(self.user)

        response = self.client.get(self.view_url)
        self.assertEqual(response.status_code, 200)


class TestSubmittedView(TestCase):
    view_name = "submitted-feedback"

    def setUp(self):
        self.view_url = reverse(self.view_name)
        self.user = UserFactory()

    def test_no_permission(self):
        self.client.force_login(self.user)

        response = self.client.get(self.view_url, follow=True)
        self.assertEqual(response.status_code, 403)

    def test_has_permission(self):
        self.user.user_permissions.add(
            Permission.objects.get(
                name="Can view feedback",
            ),
        )
        self.client.force_login(self.user)

        response = self.client.get(self.view_url)
        self.assertRedirects(response, "/submitted/default/")


class TestSubmittedListingView(TestCase):
    view_name = "feedback-listing"

    def setUp(self):
        self.view_url = reverse(self.view_name, kwargs={"form_id": "default"})
        self.user = UserFactory()

    def test_no_permission(self):
        self.client.force_login(self.user)

        response = self.client.get(self.view_url)
        self.assertEqual(response.status_code, 403)

    def test_has_permission(self):
        self.user.user_permissions.add(
            Permission.objects.get(
                name="Can view feedback",
            ),
        )
        self.client.force_login(self.user)

        response = self.client.get(self.view_url)
        self.assertEqual(response.status_code, 200)



# Generated by Django 3.2.18 on 2023-05-12 05:15

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('django_feedback_govuk', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feedback',
            name='submitter',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



import factory
from django.contrib.auth import get_user_model
from factory.django import DjangoModelFactory


User = get_user_model()


class UserFactory(DjangoModelFactory):
    class Meta:
        model = User

    username = factory.Sequence(lambda n: f"user{n}")
    email = factory.Sequence(lambda n: f"user{n}@example.com")
    password = factory.PostGenerationMethodCall("set_password", "password")
